""" Actions for events"""

from pydantic import BaseModel


class Action(BaseModel):
    """Base class for actions."""

    def get_action(self) -> None:
        raise NotImplementedError("Subclasses must implement this method.")


class EmailNotification(Action):
    """
    Action that sends an email notification.

    Attributes:
        email_address (str): The email address to send the notification to.
    """
    email_address: str

    def get_action(self):
        return {"email_address": self.email_address, "action_type": "email"}


class WebhookNotification(Action):
    """
    Action that sends a webhook notification.

    Attributes:
        webhook_url (str): The webhook url to send the notification to.
    """
    webhook_url: str

    def get_action(self):
        return {"webhook_url": self.webhook_url, "action_type": "webhook"}



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from inductiva.client.models.providers import Providers
from typing import Optional, Set
from typing_extensions import Self


class MachineType(BaseModel):
    """
    MachineType
    """

  # noqa: E501
    machine_type: StrictStr
    num_vcpus: StrictInt
    ram_gb: Union[StrictFloat, StrictInt]
    price: Union[StrictFloat, StrictInt]
    provider_id: Optional[Providers] = None
    threads_per_core: Optional[StrictInt] = None
    spot: Optional[StrictBool] = None
    region: Optional[StrictStr] = None
    zone: Optional[StrictStr] = None
    num_gpus: Optional[StrictInt] = None
    gpu_name: Optional[StrictStr] = None
    __properties: ClassVar[List[str]] = [
        "machine_type", "num_vcpus", "ram_gb", "price", "provider_id",
        "threads_per_core", "spot", "region", "zone", "num_gpus", "gpu_name"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of MachineType from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if threads_per_core (nullable) is None
        # and model_fields_set contains the field
        if self.threads_per_core is None and "threads_per_core" in self.model_fields_set:
            _dict['threads_per_core'] = None

        # set to None if spot (nullable) is None
        # and model_fields_set contains the field
        if self.spot is None and "spot" in self.model_fields_set:
            _dict['spot'] = None

        # set to None if region (nullable) is None
        # and model_fields_set contains the field
        if self.region is None and "region" in self.model_fields_set:
            _dict['region'] = None

        # set to None if zone (nullable) is None
        # and model_fields_set contains the field
        if self.zone is None and "zone" in self.model_fields_set:
            _dict['zone'] = None

        # set to None if num_gpus (nullable) is None
        # and model_fields_set contains the field
        if self.num_gpus is None and "num_gpus" in self.model_fields_set:
            _dict['num_gpus'] = None

        # set to None if gpu_name (nullable) is None
        # and model_fields_set contains the field
        if self.gpu_name is None and "gpu_name" in self.model_fields_set:
            _dict['gpu_name'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of MachineType from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "machine_type": obj.get("machine_type"),
            "num_vcpus": obj.get("num_vcpus"),
            "ram_gb": obj.get("ram_gb"),
            "price": obj.get("price"),
            "provider_id": obj.get("provider_id"),
            "threads_per_core": obj.get("threads_per_core"),
            "spot": obj.get("spot"),
            "region": obj.get("region"),
            "zone": obj.get("zone"),
            "num_gpus": obj.get("num_gpus"),
            "gpu_name": obj.get("gpu_name")
        })
        return _obj



"""List the user quotas information via CLI."""
from typing import TextIO
import argparse
import logging
import sys

from inductiva import _cli
from inductiva._cli.cmd_user import info


def get_quotas(_, fout: TextIO = sys.stdout):
    """ Lists the user's quotas.

    Lists all the user's quotas and the quotas left for the user to use.
    """
    logging.warning("\nWARNING: `inductiva quotas list` is deprecated"
                    " and will be removed in a future release.\n"
                    "Please use the `inductiva user info` command instead.\n")
    info.get_info(_, fout)
    return 0


def register(parser):
    """Register the quotas list command."""
    subparser = parser.add_parser("list",
                                  help="List the user's quotas.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva quotas list` command provides "
                             "an overview of your quotas.\n"
                             "It lists all your available quotas as well as the"
                             "quotas left for you to use\n")

    _cli.utils.add_watch_argument(subparser)

    subparser.set_defaults(func=get_quotas)



"""Util functions to handle input and output data for API requests.

This module contains several functions related to packing and unpacking
of inputs and ouputs for Web API requests.
Additionally, it contains some global constant variables defining
configurations related to paths where certain files are expected to be.
"""
import os
import pathlib
import zipfile
import tempfile
import shutil
from typing import Callable, List
from tqdm import tqdm
import fsspec
import urllib3

import logging

ARTIFACTS_DIRNAME = "artifacts"
INPUT_DIRNAME = "sim_dir"


def pack_input(input_dir, zip_name) -> str:
    """Pack all inputs into a zip file.

    Pack all input params and compress all files into a zip file.
    All required files are created in a temporary directory which is then
    compressed with "zip" format. The path of the resulting zip file is
    returned.

    Args:
        input_dir: Directory containing the input files to be uploaded.
        zip_name: Name of the zip file to be created.

    Return:
        Returns a path to zip file with the compressed input. The zip will be
        located in the temporary directory of the OS (`/tmp` in linux).
    """
    with tempfile.TemporaryDirectory() as tmpdir_path:
        # Copy input directory to a temporary directory
        dst_fullpath = os.path.join(tmpdir_path, INPUT_DIRNAME)

        if input_dir:
            shutil.copytree(input_dir, dst_fullpath)

        # Zip everything in the temporary directory into a single zip file
        zip_path = shutil.make_archive(
            os.path.join(tempfile.gettempdir(), zip_name), "zip", tmpdir_path)

        logging.debug("Compressed inputs to %s", zip_path)

    return zip_path


def extract_subdir_files(zip_fp: zipfile.ZipFile, dir_name: str,
                         output_dir: pathlib.Path):
    """Util function to extract the contents of a directory in a ZIP archive.

    For instance, if a ZIP archive contains a directory called `dir_name`,
    the contents of that directory are extracted directly to `output_dir`.

    Args:
        zip_fp: ZipFile from which to extract the directory.
        dir_name: Name of the directory inside the ZIP archive.
        output_dir: Destination directory of the contents of `data_dir`.
    """
    for member in zip_fp.namelist():
        is_dir = not os.path.basename(member)

        if not member.startswith(dir_name) or is_dir:
            continue

        src_file = zip_fp.open(member)
        target_relative_path = pathlib.Path(member).relative_to(dir_name)
        target_path = os.path.join(output_dir, target_relative_path)

        os.makedirs(os.path.dirname(target_path), exist_ok=True)

        with open(target_path, "wb") as f:
            shutil.copyfileobj(src_file, f)


def zip_dir(dir_path, zip_name):
    """Compress a directory into a zip file."""
    zip_path = shutil.make_archive(
        os.path.join(tempfile.gettempdir(), zip_name), "zip", dir_path)

    logging.debug("Compressed inputs to %s", zip_path)

    return zip_path


def _extract_zip_file_to_dir(
    dest_dir: pathlib.Path,
    remove_zip_file: zipfile.ZipFile,
    filename: str,
    zip_path: str,
):
    """Write a file from a ZIP archive to the output directory.

    Args:
        dest_dir: Directory where to store the extracted file.
        remove_zip_file: ZipFile object from which to extract the file.
        filename: Name of the file to extract.
        zip_path: Path of the file inside the ZIP archive.
    """
    with remove_zip_file.open(zip_path) as source:
        target_path = dest_dir / pathlib.Path(filename)
        target_path.parent.mkdir(parents=True, exist_ok=True)
        with open(target_path, "wb") as target:
            target.write(source.read())


def _download_partial_files(
    download_url: str,
    filenames: List[str],
    dest_dir: pathlib.Path,
    make_zip_path: Callable,
) -> None:
    """Download the partial files of a task.

    Args:
        download_url: URL from which to download the files.
        filenames: List of filenames to download.
        dest_dir: Path where to store the downloaded files.
        make_zip_path: Function to create a path of a file inside the ZIP 
        given a filename.

    Return:
        Returns True if the download was successful, False otherwise.
    """
    try:
        remote_filesystem = fsspec.filesystem("http")

        with remote_filesystem.open(download_url, "rb") as remote_file:
            with zipfile.ZipFile(remote_file) as remote_zip_file:
                for filename in filenames:
                    zip_path = make_zip_path(filename)
                    try:
                        _extract_zip_file_to_dir(dest_dir, remote_zip_file,
                                                 filename, zip_path)
                    except KeyError:
                        logging.warning(
                            "File %s not found in the output archive.",
                            zip_path)

    except Exception as e:  # pylint: disable=broad-except
        logging.debug("Error downloading partial outputs: %s", e)
        logging.error("Partial download failed.")


def download_partial_outputs(
    download_url: str,
    filenames: List[str],
    output_dir: pathlib.Path,
) -> None:
    """Download the partial outputs of a task.

    Args:
        download_url: URL from which to download the outputs.
        filenames: List of filenames to download.
        output_dir: Path where to store the downloaded files.

    Return:
        Returns True if the download was successful, False otherwise.
    """
    return _download_partial_files(
        download_url=download_url,
        filenames=filenames,
        dest_dir=output_dir,
        make_zip_path=lambda filename: "artifacts/" + filename)


def download_partial_inputs(
    download_url: str,
    filenames: List[str],
    input_dir: pathlib.Path,
) -> None:
    """Download the partial inputs of a task.

    Args:
        download_url: URL from which to download the inputs.
        filenames: List of filenames to download.
        output_dir: Path where to store the downloaded files.

    Return:
        Returns True if the download was successful, False otherwise.
    """
    return _download_partial_files(
        download_url=download_url,
        filenames=filenames,
        dest_dir=input_dir,
        make_zip_path=lambda filename: "sim_dir/" + filename \
            if filename != "input.json" else filename)


def download_file(
    response: urllib3.response.HTTPResponse,
    output_path: pathlib.Path,
    chunk_size: int = 1000,
) -> None:
    """Download a file from a urllib3 response object.

    Use a urllib3 response object to download a file, showing a progress bar.

    Args:
        response: urllib3 response object.
        output_path: Path where to store the downloaded file.
        chunk_size: Size of the chunks in which to download the file.
    """
    # if the response header does not contain a x-content-length header,
    # the progress bar will not be displayed correctly, but download
    # will still work.
    # "x-content-length" is a custom header that is set by the API instead
    # of the standard "content-length" header, because the API needs to use
    # "transfer-encoding: chunked" to stream the response.
    # If the download is provided by a file server, the "content-length"
    # header will be used.
    download_size = response.headers.get(
        "x-content-length") or response.headers.get("content-length", 0)

    with tqdm(
            total=int(download_size),
            unit="B",
            unit_scale=True,
            unit_divisor=1000,  # Use 1 KB = 1000 bytes
    ) as progress_bar:
        with open(output_path, "wb") as f:
            while chunk := response.read(chunk_size):
                f.write(chunk)
                progress_bar.update(len(chunk))

    response.release_conn()


def decompress_zip(zip_path: pathlib.Path, output_dir: pathlib.Path):
    """Uncompress a ZIP archive containing the outputs of a task.

    If the archive contains the directory called artifacts, it means that the
    download includes the full outputs of the task with the full directory
    structure of the outputs (output.json, artifacts/*). Only the contents
    inside artifacts are extracted to the output directory. If the archive
    does not contain the artifacts directory, it means that the download
    only includes a few files, which are directly in the root of the archive,
    without the `artifacts` directory.
    """
    with zipfile.ZipFile(zip_path, "r") as zip_f:
        full_output = True
        try:
            zip_f.getinfo("artifacts/")
        except KeyError:
            full_output = False

        if full_output:
            extract_subdir_files(
                zip_f,
                "artifacts",
                output_dir,
            )
        else:
            zip_f.extractall(output_dir)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import StrictStr
from inductiva.client.models.backend_version import BackendVersion
from inductiva.client.models.version_comparison_result import VersionComparisonResult

from inductiva.client.api_client import ApiClient, RequestSerialized
from inductiva.client.api_response import ApiResponse
from inductiva.client.rest import RESTResponseType


class VersionApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_call
    def compare_client_and_backend_versions(
        self,
        client_version: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> VersionComparisonResult:
        """Compare Client And Backend Versions

        Endpoint to compare the client and server version numbers.  Args: - client_version (str): The version string sent by the client in                         'MAJOR.MINOR.PATCH' format.  Raises: - HTTPException: 400 Bad Request if the client_version is not                  formatted properly. - HTTPException: 406 Not acceptable if the client_version is not                  supported by the server.  Returns: - VersionComparisonResult: A model indicating whether the client version                            matches the server version, along with the                            provided client version, the current server                            and the minimum supported client version.

        :param client_version: (required)
        :type client_version: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._compare_client_and_backend_versions_serialize(
            client_version=client_version,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "VersionComparisonResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def compare_client_and_backend_versions_with_http_info(
        self,
        client_version: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[VersionComparisonResult]:
        """Compare Client And Backend Versions

        Endpoint to compare the client and server version numbers.  Args: - client_version (str): The version string sent by the client in                         'MAJOR.MINOR.PATCH' format.  Raises: - HTTPException: 400 Bad Request if the client_version is not                  formatted properly. - HTTPException: 406 Not acceptable if the client_version is not                  supported by the server.  Returns: - VersionComparisonResult: A model indicating whether the client version                            matches the server version, along with the                            provided client version, the current server                            and the minimum supported client version.

        :param client_version: (required)
        :type client_version: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._compare_client_and_backend_versions_serialize(
            client_version=client_version,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "VersionComparisonResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def compare_client_and_backend_versions_without_preload_content(
        self,
        client_version: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Compare Client And Backend Versions

        Endpoint to compare the client and server version numbers.  Args: - client_version (str): The version string sent by the client in                         'MAJOR.MINOR.PATCH' format.  Raises: - HTTPException: 400 Bad Request if the client_version is not                  formatted properly. - HTTPException: 406 Not acceptable if the client_version is not                  supported by the server.  Returns: - VersionComparisonResult: A model indicating whether the client version                            matches the server version, along with the                            provided client version, the current server                            and the minimum supported client version.

        :param client_version: (required)
        :type client_version: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._compare_client_and_backend_versions_serialize(
            client_version=client_version,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "VersionComparisonResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _compare_client_and_backend_versions_serialize(
        self,
        client_version,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if client_version is not None:

            _query_params.append(('client_version', client_version))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = []

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/version-check',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def get_backend_version(
        self,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> BackendVersion:
        """Get Backend Version

        Returns the latest version of the api

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_backend_version_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BackendVersion",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_backend_version_with_http_info(
        self,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[BackendVersion]:
        """Get Backend Version

        Returns the latest version of the api

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_backend_version_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BackendVersion",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_backend_version_without_preload_content(
        self,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get Backend Version

        Returns the latest version of the api

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_backend_version_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BackendVersion",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _get_backend_version_serialize(
        self,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = []

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/version',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)



"""MOHID simulator module of the API."""
from typing import Optional, Union

from inductiva import simulators, tasks, types

ALLOWD_MOHID_COMMANDS = ["MohidWater.exe", "MohidLand.exe"]


@simulators.simulator.mpi_enabled
class MOHID(simulators.Simulator):
    """Class to invoke a generic MOHID simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the MOHID simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "mohid"

    def run(self,
            input_dir: Optional[str],
            *,
            command: str = "MohidWater.exe",
            working_dir: str = "",
            on: types.ComputationalResources,
            run_ddc: bool = False,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            command: MOHID command to run. Can be `MohidWater.exe` or
                `MohidLand.exe`.
            working_dir: Path (relative to the input directory) to the directory
                where the simulation will run. The command picked and
                domain consolidation will run on that folder.
            run_ddc: Boolean that will determine if we are going to run 
                domain consolidation (`MohidDDC.exe`) after the
                simulation ends. This will combine the results of the multiple
                MPI processes into one.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """

        self._check_vcpus(n_vcpus, on)

        if command not in ALLOWD_MOHID_COMMANDS:
            raise ValueError(
                f"Command {command} not supported.\n"
                f"Available commands are: {ALLOWD_MOHID_COMMANDS}.")

        #We are using the intel MPI from oneapi
        #We don't have support for this MPI on the back end yet, this is why
        #we are not creating an MPICONFIG.
        mpi_command = "mpirun "
        if n_vcpus:
            mpi_command += f"-np {n_vcpus} "

        commands = [f"{mpi_command} /usr/local/bin/{command}"]

        if run_ddc:
            commands.append("/usr/local/bin/MohidDDC.exe")

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           remote_assets=remote_assets,
                           run_subprocess_dir=working_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Register CLI commands for logs."""
import argparse
import os

from .. import loader, utils
from ... import constants


def register(root_parser):
    parser = root_parser.add_parser(
        "auth",
        help="Authentication commands",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.description = ("Authentication management utilities.\n\n"
                          "The `inductiva auth` command allows you to "
                          "manage user authentication.\n")

    utils.show_help_msg(parser)

    subparsers = parser.add_subparsers(title="available subcomands")
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         package=__name__,
                         ignores_prefix=constants.LOADER_IGNORE_PREFIX)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictStr
from typing import Any, ClassVar, Dict, List
from inductiva.client.models.action1 import Action1
from inductiva.client.models.trigger1 import Trigger1
from typing import Optional, Set
from typing_extensions import Self


class EventInfo(BaseModel):
    """
    EventInfo
    """

  # noqa: E501
    event_id: StrictStr
    trigger: Trigger1
    action: Action1
    __properties: ClassVar[List[str]] = ["event_id", "trigger", "action"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of EventInfo from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of trigger
        if self.trigger:
            _dict['trigger'] = self.trigger.to_dict()
        # override the default output from pydantic by calling `to_dict()` of action
        if self.action:
            _dict['action'] = self.action.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of EventInfo from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "event_id":
                obj.get("event_id"),
            "trigger":
                Trigger1.from_dict(obj["trigger"])
                if obj.get("trigger") is not None else None,
            "action":
                Action1.from_dict(obj["action"])
                if obj.get("action") is not None else None
        })
        return _obj



"""CLI utils."""
import argparse
import os

import inductiva
from inductiva import constants

BEGIN_TAG = "# >>> INDUCTIVA BEGIN:"
END_TAG = "# >>> INDUCTIVA END:"


def positive_float(value):
    """Check if the given value is a positive float."""
    try:
        value = float(value)
    except TypeError as _:
        raise argparse.ArgumentTypeError(f"{value} is an invalid type.")
    if value <= 0:
        raise argparse.ArgumentTypeError(f"{value} must be a positive number.")
    return value


def add_watch_argument(subparser):
    """Add watch flag to the subparser that prompts the commands every Nsecs."""
    subparser.set_defaults(watchable=True)
    subparser.add_argument("-w",
                           "--watch",
                           nargs="?",
                           const=2.0,
                           type=positive_float,
                           help="Prompt the command every N seconds.")


def check_running_for_first_time():
    """Checks if it is the first time the cli is being called.

    It does this by checking if there is a folder:

      `$HOME/.inductiva/v{inductiva_version}`

    If there is not returns true and creates that folder.

    """
    version = inductiva.__version__
    dir_name = constants.HOME_DIR / f"v{version}"
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)
        return True
    return False


def show_help_msg(parser):
    """Show help message for command if no subcommand is provided."""
    parser.set_defaults(func=lambda _: parser.print_help())



"""Helper functions for the templating module."""
import os

TEMPLATE_EXTENSION = ".jinja"


def is_template(file: str) -> bool:
    """Check if the given file is of template type."""
    return file.endswith(TEMPLATE_EXTENSION)


def strip_extension(file: str) -> str:
    """Strip the template extension if the given file is a template."""
    if not is_template(file):
        return file
    return os.path.splitext(file)[0]


def get_dir_structure(template_dir: str):
    """
    Generate a dictionary with the structure of the template directory.
    The keys are the subdirectories of the template directory, relative to
    the template directory itself. The values are dictionaries with the keys
    "files" and "templates" containing the lists of the files and templates
    in each subdirectory, respectively.
    Example:
    >>> renderer.get_dir_structure("template_dir")
    {
        ".": {
            "files": ["file0.1"],
            "templates": ["template0.1"]
        },
        "subdir1": {
            "files": ["file1.1", "file1.2"],
            "templates": ["template1.1", "template1.2"]
        }
        ...
    }
    """
    return {
        os.path.relpath(root, start=template_dir): {
            "templates": [f for f in files if is_template(f)],
            "files": [f for f in files if not is_template(f)],
        } for root, _, files in os.walk(template_dir, topdown=True)
    }



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from inductiva.client.models.operation_status import OperationStatus
from inductiva.client.models.storage_operation_name import StorageOperationName
from typing import Optional, Set
from typing_extensions import Self


class StorageOperation(BaseModel):
    """
    StorageOperation
    """

  # noqa: E501
    id: StrictStr
    name: StorageOperationName
    status: OperationStatus
    error_message: Optional[StrictStr] = None
    attributes: Dict[str, Any]
    start_time: datetime
    end_time: Optional[datetime] = None
    __properties: ClassVar[List[str]] = [
        "id", "name", "status", "error_message", "attributes", "start_time",
        "end_time"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of StorageOperation from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if error_message (nullable) is None
        # and model_fields_set contains the field
        if self.error_message is None and "error_message" in self.model_fields_set:
            _dict['error_message'] = None

        # set to None if end_time (nullable) is None
        # and model_fields_set contains the field
        if self.end_time is None and "end_time" in self.model_fields_set:
            _dict['end_time'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of StorageOperation from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "name": obj.get("name"),
            "status": obj.get("status"),
            "error_message": obj.get("error_message"),
            "attributes": obj.get("attributes"),
            "start_time": obj.get("start_time"),
            "end_time": obj.get("end_time")
        })
        return _obj



"""COAWST simulator module of the API."""

import os
from typing import List, Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class COAWST(simulators.Simulator):
    """Class to invoke a generic COAWST simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the COAWST simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "coawst"

    def _validate_build_script(self, build_script: str) -> None:
        """
        Validates the build script by checking if required environment variables
        are correctly set.

        :param build_script: Path to the build script file.
        :raises ValueError: If any required variable is missing or incorrect.
        """
        checks = {
            #Check if MY_ROOT_DIR points to the correct folder
            r"export\s+MY_ROOT_DIR\s*=\s*/workdir/output/artifacts/__COAWST":
                "Wrong value for MY_ROOT_DIR in your build_coawst_script.\n"
                "MY_ROOT_DIR needs to be /workdir/output/artifacts/__COAWST",
            #Check if USE_MPI is on
            r"export\s+USE_MPI\s*=\s*on":
                "Wrong value for USE_MPI in your build_coawst_script.\n"
                "USE_MPI needs to be on",
            #Check if which_MPI is set to openmpi
            r"export\s+which_MPI\s*=\s*openmpi":
                "Wrong value for which_MPI in your build_coawst_script.\n"
                "which_MPI needs to be openmpi"
        }

        for regex, error_msg in checks.items():
            if not self._regex_exists_in_file(build_script, regex):
                raise ValueError(error_msg)

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            *,
            build_coawst_script: Optional[str] = None,
            on: types.ComputationalResources,
            coawst_bin: str = "coawstM",
            init_commands: Optional[List[str]] = None,
            cleanup_commands: Optional[List[str]] = None,
            compile_simulator: bool = True,
            use_hwthread: bool = True,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: Name of the simulation configuration file.
            build_coawst_script: Script used to build the COAWST executable.
            coawst_bin : str
                Name of the COAWST binary to execute (defaults to 'coawstM').
                This binary will be used whether the simulator is compiled or
                not. If compilation is skipped, the binary must
                already exist in the `input_dir`.
            init_commands: List of helper commands to prepare things for your
                simulation. Used to copy files to and from
                `/workdir/output/artifacts/__COAWST`. It can also be used
                to run any helper function present within COAWST, like
                `scrip_coawst`.
                Will run before the compilation of the simulator.
            cleanup_commands: List of helper commands to clean up things
                after your simulation finishes. Used to copy files from
                `/workdir/output/artifacts/__COAWST` to your input_dir. Or
                to delete unwanted files generated during the simulation.
                Will run after the simulation ends.
            compile_simulator :
                If True, the simulator will be compiled using the provided
                `build_coawst_script`, and the simulation will run using the
                specified `coawst_bin`.  
                If False, the simulation will be run directly using the
                precompiled `coawst_bin` binary, which should already exist in
                the `input_dir` with the same name as `coawst_bin`.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """

        if build_coawst_script is None and compile_simulator:
            raise ValueError("build_coawst_script is required if "
                             "compile_simulator is True")

        self._check_vcpus(n_vcpus, on)

        #only runs checks if we dont use remote assets

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename=sim_config_filename)

        # only validates the build script if we are going to compile
        if compile_simulator:
            self._input_files_exist(input_dir=input_dir,
                                    remote_assets=remote_assets,
                                    build_coawst_script=build_coawst_script)

            build_script = os.path.join(input_dir, build_coawst_script)
            self._validate_build_script(build_script=str(build_script))

        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus
        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)

        #only compiles if we are not using a precompiled binary
        if compile_simulator:
            # 34 selects dmpar for linux when compiling WRF
            compilation_command = [Command(f"bash {build_coawst_script}", "34")]
        else:
            # Copy the simulation binary to the __COAWST directory
            # and make it executable
            compilation_command = [
                f"cp -f {coawst_bin} __COAWST/",
                f"chmod +x __COAWST/{coawst_bin}",
            ]

        commands = [
            #Copy COAWST source code to our input dir
            "cp -r /opt/COAWST /workdir/output/artifacts/__COAWST",
            "create_all_sim_links",
            #Either compiles the simulator or copies the provided binary
            *compilation_command,
            #Run the simulation
            Command(f"{coawst_bin} {sim_config_filename}",
                    mpi_config=mpi_config),
            #Cleanup COAWST source code
            # and remove the symlinks
            "rm -r  __COAWST",
            "clean_all_sim_links"
        ]

        # Add init commands after building and before running the simulation
        if init_commands is not None:
            commands = commands[:1] + init_commands + commands[1:]

        # Add cleanup commands after running the simulation
        if cleanup_commands is not None:
            # add comands to the penultimate position
            commands = commands[:-2] + cleanup_commands + commands[-2:]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           remote_assets=remote_assets,
                           resubmit_on_preemption=resubmit_on_preemption,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Methods to interact with the tasks submitted to the API."""
from collections import defaultdict
from typing import Any, Iterable, List, Mapping, Optional

import inductiva
import inductiva.client
from inductiva.client import models
from inductiva.tasks.task import Task
from inductiva.utils import format_utils
from inductiva.client import ApiException


def to_dict(list_of_tasks: Iterable[Task]) -> Mapping[str, List[Any]]:
    """
    Converts an Iterable of tasks to a dictionary with all the
    relevant information for all the tasks.
        Args:
            list_of_tasks: An Iterable of tasks.
        Returns:
            A dictionary with all the relevant information for
            all the tasks. Example: { "ID": [1, 2, 3],
            "Simulator": ["reef3d", "reef3d", "reef3d"], ... }
    """
    column_names = [
        "ID", "Simulator", "Status", "Submitted", "Started", "Computation Time",
        "Resource Type"
    ]
    table = defaultdict(list, {key: [] for key in column_names})

    for task in list_of_tasks:
        execution_time = task.get_computation_time(cached=True)

        if execution_time is not None:
            execution_time = format_utils.seconds_formatter(execution_time)
            if task.info.computation_end_time is None:
                if task.info.status in ["started", "submitted"]:
                    execution_time = f"*{execution_time}"
                else:
                    execution_time = "n/a"

        if task.info.executer is None:
            resource_type = None
        else:
            if task.info.executer.vm_type == "n/a":
                vm_type = task.info.executer.vm_name
            else:
                vm_type = task.info.executer.vm_type
            resource_type = (f"{task.info.executer.host_type} "
                             f"{vm_type}")
            if task.info.executer.n_mpi_hosts > 1:
                resource_type += f" x{task.info.executer.n_mpi_hosts}"

        table["ID"].append(task.id)
        table["Simulator"].append(task.get_simulator_name())
        table["Status"].append(task.info.status_alias)
        table["Submitted"].append(task.info.input_submit_time)
        table["Started"].append(task.info.start_time)
        table["Computation Time"].append(execution_time)
        table["Resource Type"].append(resource_type)

    return table


def _fetch_tasks_from_api(status: Optional[str] = None,
                          page=1,
                          per_page=10,
                          project: Optional[str] = None) -> List[models.Task]:
    """Get information about a user's tasks on the API.

    Tags can be filtered by a status. Results are paginated indexed from 1.
    """

    if status is not None:
        status = models.TaskStatusCode(status)

    with inductiva.api.methods.get_client() as client:
        api_instance = inductiva.client.TasksApi(client)
        try:
            # Get User Tasks
            response = api_instance.get_user_tasks(
                page=page,
                per_page=per_page,
                project=project,
                status=status,
            )

            return response.tasks

        except ApiException as e:
            raise e


def get(last_n: int = 5,
        status: Optional[str] = None,
        project: Optional[str] = None) -> List["inductiva.tasks.Task"]:
    """Get the last N tasks of a user.

    This function fetches info about the last N tasks (with respect to
    submission time) of a user to stdout, sorted by submission time with the
    most recent first.
    A status can be specified to filter to get only tasks with that status, in
    which case the last N tasks with that status will be listed.
    The number of tasks can be less than N if the aren't enough tasks that match
    the specified criteria.

    Similar to the inductiva.task.list() function, but instead of printing
    to stdout, returns a list of Task objects which can be used to perform
    further operations on those tasks (e.g., download outputs,
    kill unfinished tasks, etc.).

    Example usage:
        # get the last 5 tasks that haven't started yet and kill them
        tasks = inductiva.tasks.get(5, status="submitted")
        for task in tasks:
            task.kill()

    Args:
        last_n: The number of most recent tasks with respect to submission
            time to fetch. If filtering criteria (currently status is available)
            is specified, most recent N tasks that match that criteria will be
            listed. The actual number of tasks may be less if there
            aren't enough tasks available.
        status: The status of the tasks to get. If None, tasks with any status
            will be returned.
        project: The project from which to fetch. If None, fetches from all
            projects.

    Returns:
        List of Task objects.
    """
    if last_n < 1:
        raise ValueError("last_n must be >= 1")

    raw_tasks_info = _fetch_tasks_from_api(status,
                                           page=1,
                                           per_page=last_n,
                                           project=project)
    tasks = [
        inductiva.tasks.Task.from_api_info(info) for info in raw_tasks_info
    ]

    return tasks


def get_tasks(last_n: int = 10,
              project: Optional[str] = None,
              status: Optional[str] = None):
    """Get the last N submitted tasks.

        Get the last N submitted tasks, eventually filtered by status.
        By default, only the last 10 submitted tasks are returned,
        irrespectively of their status.

        Args:
            last_n (int): The number of tasks with repect to the submission
                time to fectch. If `last_n<=0` we fetch all tasks submitted
                to the project.
            status: Status of the tasks to get. If `None`, tasks with any
                status will be returned.
        """
    if last_n <= 0:
        return inductiva.tasks.get_all(status=status, project=project)
    return inductiva.tasks.get(last_n=last_n, status=status, project=project)


def get_all(
    status: Optional[str] = None,
    project: Optional[str] = None,
) -> List["inductiva.tasks.Task"]:
    """Get all tasks of a user.

    This function fetches all tasks of a user, sorted by submission
    time with the most recent first. If status is specified, only
    tasks with that status will be fetched.
    Args:
        status: The status of the tasks to get. If None, tasks with any status
            will be returned.
    Returns:
        List of dictionaries with information about the tasks.
    """
    all_tasks = []
    page_counter = 1
    while tasks_fetched := _fetch_tasks_from_api(status,
                                                 page=page_counter,
                                                 per_page=500,
                                                 project=project):
        all_tasks.extend(tasks_fetched)
        page_counter += 1

    tasks = [inductiva.tasks.Task.from_api_info(info) for info in all_tasks]

    return tasks



"""Class to run commands on an custom image."""
from typing import List, Optional, Union

from inductiva import simulators, tasks, types


@simulators.simulator.mpi_enabled
class CustomImage(simulators.Simulator):
    """Class to run commands on an custom image."""

    def __init__(self, container_image: str):
        """Initialize the ArbitraryImage class.
        Point to the API method to run a simulation.
        Args:
            container_image: The container image to use for the simulation.
                Example: container_image="docker://inductiva/kutu:xbeach_v1.23"
        """
        self.container_image = container_image
        super().__init__()

        self.simulator = "arbitrary_commands"

    def _get_image_uri(self):
        return self.container_image

    def run(self,
            input_dir: Optional[str],
            commands: List[str],
            *,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.
        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run.
            on: The computational resource to launch the simulation on.
            storage_dir: Parent directory for storing simulation
                               results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           container_image=self._image_uri,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Top command for task."""
from typing import TextIO
import argparse
import sys

from inductiva import tasks


def last_modifed_file(args: argparse.Namespace, fout: TextIO = sys.stdout):
    """
    Display the last modified file for a given task.

    This function retrieves and streams information about the most recently 
    modified file associated with a specified task. It validates that the 
    task computation has started before proceeding. If the task is invalid 
    or not started, an error message is printed to `stderr`.
    """
    task_id = args.id
    task = tasks.Task(task_id)

    result, return_code = task.last_modified_file()

    if result:
        print(result, file=fout)

    return return_code


def register(parser):
    """Register the last-modifed-file command."""
    subparser = parser.add_parser("last-modified-file",
                                  help="Displays the last modified file.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva tasks last-modified-file` command displays the last"
        "modified file for the simulation. It also shows the last time the file"
        "was modified and how long it has been since that.")

    subparser.add_argument("id",
                           type=str,
                           help="The ID of the task to get the last modifed "
                           "file.")

    subparser.set_defaults(func=last_modifed_file)



"""A simple ANSI-based pager for displaying formatted text in a terminal.

The module provides a file-like object that can be used to display text in a
terminal. The text is displayed in a paged manner, allowing the user to scroll
up and down, and to quit the display.
The pager requires the terminal to support ANSI escape codes.
"""
from typing import Optional, Type, Iterator, AnyStr, Iterable
from collections import namedtuple
from types import TracebackType
from functools import partial
import threading
import termios
import typing
import shutil
import tty
import sys
import re
import io
import os

from inductiva.utils import scheduler

strip_escaped = partial(re.compile(r"\x1b\[[;\d]*[A-Za-z]").sub, "")
strip_escaped.__doc__ = "Return the string with all ANSI escape codes removed."


def get_output_size():
    """Return the size of the terminal window as a named tuple."""
    return shutil.get_terminal_size()


Pad = namedtuple("Pad", "width height column line")
Pad.__doc__ = """A named tuple representing a pad in the terminal.
the tuple has the following fields:
- width: The width of the pad.
- height: The height of the pad.
- column: The column where the pad starts.
- line: The line where the pad starts.
"""


class Display:
    """A display object that holds the pads for the header, body and footer."""
    __slots__ = ["window", "header", "body", "footer"]

    def __init__(self):
        """Initialize the display object."""
        self.set_size(*get_output_size())

    def set_size(self, width: int, height: int) -> None:
        """Set the size of the display.
        Sets the size of the window and updates the sizes of the
        header, body and footer pads.

        Args:
            width (int): The width of the display.
            height (int): The height of the display.
        """
        self.window = Pad(width, height, 1, 1)
        self.header = Pad(width, 1, 1, 1)
        self.body = Pad(width, height - 2, 1, 2)
        self.footer = Pad(width, 1, 1, height)

    def resized(self) -> bool:
        """Return whether the terminal has been resized.

        If the terminal has been resized, the size of the display is updated and
        all the pads are recalculated. The method returns True if the terminal
        has been resized, False otherwise.
        """
        size = get_output_size()
        resized = size != (self.window.width, self.window.height)
        if resized:
            self.set_size(*size)
        return resized


class PagedOutput(io.TextIOBase):
    """A file-like object for displaying text in a terminal.

    The object provides a simple interface for displaying text in a terminal
    in a paged manner. The text is displayed in a buffer and the user can
    navigate through the buffer using the arrow keys. The object is also
    capable of displaying a header and a footer, both of which can be configured
    by the user to display custom content.
    """
    DEFAULT_FOOTER = "Type q to quit, ← ↑ → ↓ to navigate."

    SCROLL_DOWN = 1
    SCROLL_HOME = 0
    SCROLL_UP = -1

    def __init__(self, header: str = None, footer: str = DEFAULT_FOOTER):
        self._lock = threading.RLock()
        self._display = Display()
        self._scrolled_to = 0
        self._resizer = None
        self._header = None
        self._footer = None
        self._buffer = []

        self._enter_alt_screen()
        self.set_header(header)
        self.set_footer(footer)

        self._resizer = scheduler.StoppableScheduler(0.5, self._check_resize)
        self._resizer.daemon = True
        self._resizer.start()

    def _check_resize(self) -> None:
        """Check if the terminal has been resized and update the display."""
        with self._lock:
            if self._display.resized():
                self._render_header()
                self._render_body(self._scrolled_to)
                self._render_footer()

    def _enter_alt_screen(self) -> None:
        """Enter alternate screen and hide cursor."""
        sys.stdout.write("\033[?1049h\033[?25l")

    def _exit_alt_screen(self) -> None:
        """Exit alternate screen and show cursor."""
        sys.stdout.write("\033[?1049l\033[?25h")

    def _moveto(self, pad: Pad) -> None:
        """Move cursor to the start of the given pad."""
        sys.stdout.write(f"\033[{pad.line};{pad.column}H")

    def _clear_line(self) -> None:
        """Clear line where cursor is located."""
        sys.stdout.write("\033[2K")

    def _clear_screen(self) -> None:
        """Clear screen from cursor position till the end."""
        sys.stdout.write("\033[0J")

    def _render_header(self) -> None:
        """Render the header onto the display.

        Renders the content of the _header attribute to the header line
        onto the display.
        """
        self._moveto(self._display.header)
        self._clear_line()
        if self._header:
            sys.stdout.write(self._header)
        sys.stdout.flush()

    def _render_body(self, start: int) -> None:
        """Render the content of the text pad onto the display.

        Renders the content of the buffer, starting from the given line.
        """
        with self._lock:
            if start > self._scrolled_to + self._display.body.height:
                return

            self._moveto(self._display.body)
            self._clear_screen()
            start = self._scrolled_to
            end = start + self._display.body.height
            for line in self._buffer[start:end]:
                sys.stdout.write(line + "\n")

            sys.stdout.flush()

    def _render_footer(self) -> None:
        """Render the footer onto the display.

        Renders the content of the _footer + line tracking information
        to the footer line on the display
        """
        line_count = len(self._buffer)
        curr_line = self._scrolled_to
        if line_count:
            curr_line += 1

        left = self._footer if self._footer else ""
        stripped_left = strip_escaped(left)
        right = f"{curr_line}/{line_count}"
        fill = self._display.footer.width - len(stripped_left) - len(right)

        self._moveto(self._display.footer)
        self._clear_line()
        sys.stdout.write(left + fill * " " + right)
        sys.stdout.flush()

    def _scroll(self, direction: int, page: bool = False) -> None:
        """Scroll the content of the text pad in the given direction.

        This method scrolls a single line in the given direction when
        `page` is False. Otherwise, an entire page is scrolled.
        """
        with self._lock:
            line = self._scrolled_to
            if page:
                line += direction * self._display.body.height
            else:
                line += direction
            self.scroll_to(line)

    def set_header(self, header: str) -> None:
        """Set the header of the display.

        Args:
            header (str): The header to be displayed.
        """
        with self._lock:
            self._header = header
            self._render_header()

    def set_footer(self, footer: str) -> None:
        """Set the footer of the display.

        Args:
            footer (str): The footer to be displayed."""
        with self._lock:
            self._footer = footer if footer is not None else self.DEFAULT_FOOTER
            self._render_footer()

    def scroll_to(self, line: int) -> None:
        """Scroll the display to the given line.

        The method scrolls the display so that the given line is at the top
        of the text pad. The line is clamped to the range [0, line_count-1].
        """
        with self._lock:
            scrolled_to = min(len(self._buffer) - 1, max(0, line))
            self._scrolled_to = max(0, scrolled_to)
            self._render_body(self._scrolled_to)
            self._render_footer()

    def clear(self) -> None:
        """Clear the content of the file-like object.

        Both the text and footer pads are cleared and the display refreshed.
        """
        with self._lock:
            self._scrolled_to = 0
            self._buffer.clear()
            self._moveto(self._display.body)
            self._clear_screen()
            self._render_footer()

    def run(self) -> None:
        """Run the navigation loop"""

        stdin = sys.stdin.fileno()
        tattr = termios.tcgetattr(stdin)
        tty.setcbreak(stdin, termios.TCSANOW)

        scroll = {
            "\033[A": partial(self._scroll, self.SCROLL_UP, False),
            "\033[B": partial(self._scroll, self.SCROLL_DOWN, False),
            "\033[C": partial(self._scroll, self.SCROLL_DOWN, True),
            "\033[D": partial(self._scroll, self.SCROLL_UP, True),
            " ": partial(self._scroll, self.SCROLL_DOWN, True),
            "b": partial(self._scroll, self.SCROLL_UP, True),
        }

        try:
            buf = code = ""
            while True:
                buf = sys.stdin.read(1)
                if buf == "q":
                    break
                if action := scroll.get(buf):
                    action()
                elif buf == "\033":
                    code = buf
                elif buf in "[ABCD":
                    code += buf
                else:
                    code = ""
                    continue

                if action := scroll.get(code):
                    action()
                    code = ""

        except KeyboardInterrupt:
            termios.tcsetattr(stdin, termios.TCSANOW, tattr)

    def write(self, text: str) -> int:
        """Write string to the file-like object.

        Returns the total number of lines in the display.
        Note: The display is only updated after a call to the `flush` method.
        """
        with self._lock:
            start = len(self._buffer)
            lines = text.splitlines()
            self._buffer.extend(lines)
            self._render_header()
            self._render_body(start)
            self._render_footer()
            return len(self._buffer) - start

    def writelines(self, lines: Iterable[AnyStr]) -> None:
        """Write a list of lines to the file-like object.

        Line separators are not added, so it is usual for each of the
        lines provided to have a line separator at the end.
        Note: The display is only updated after a call to the `flush` method.
        """
        with self._lock:
            start = len(self._buffer)
            self._buffer.extend(x for line in lines for x in line.splitlines())
            self._render_body(start)
            self._render_footer()

    def flush(self) -> None:
        """Flush the content of the file-like object to the display."""
        sys.stdout.flush()

    def close(self) -> None:
        """Close the file-like object.

        The method closes the file-like object and resets the terminal
        to its previous state."""
        self._resizer.stop()
        self._resizer.join()
        self._exit_alt_screen()
        os.system("stty sane")  # Required to return terminal to normal.

    def __enter__(self) -> "PagedOutput":
        return self

    def __exit__(self, unused_t: Optional[Type[BaseException]],
                 unused_value: Optional[BaseException],
                 unused_traceback: Optional[TracebackType]) -> Optional[bool]:
        self.close()

    def isatty(self) -> bool:
        """Return whether this is an 'interactive' stream.

        Always returns False.
        """
        return False

    def writable(self) -> bool:
        """Return whether the object can be written to.

        Always returns True.
        """
        return True

    def readable(self) -> bool:
        """Return whether the object can be read from.

        Always returns False.
        """
        return False

    def seekable(self) -> bool:
        """Return whether the object supports random access.

        Always returns False.
        """
        return False

    def fileno(self) -> int:
        raise io.UnsupportedOperation("Underlying file object has no fileno")

    def read(self, n: int = ...) -> AnyStr:
        raise io.UnsupportedOperation("Underlying file object is not readable")

    def readline(self, limit: int = ...) -> AnyStr:
        raise io.UnsupportedOperation("Underlying file object is not readable")

    def readlines(self, hint: int = ...) -> typing.List[AnyStr]:
        raise io.UnsupportedOperation("Underlying file object is not readable")

    def seek(self, offset: int, whence: int = ...) -> int:
        raise io.UnsupportedOperation("File-like object is not seekable")

    def tell(self) -> int:
        raise io.UnsupportedOperation(
            "File-like object does not support the tell method")

    def truncate(self, size: Optional[int] = ...) -> int:
        raise io.UnsupportedOperation(
            "File-like object does not support the truncate method")

    def __next__(self) -> AnyStr:
        raise TypeError("File-like object is not an iterator")

    def __iter__(self) -> Iterator[AnyStr]:
        raise TypeError("File-like object is not iterable")



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from inductiva.client.models.compression_method import CompressionMethod
from inductiva.client.models.executer import Executer
from inductiva.client.models.task_machine_operation import TaskMachineOperation
from inductiva.client.models.task_metrics import TaskMetrics
from inductiva.client.models.task_status_code import TaskStatusCode
from inductiva.client.models.task_status_info import TaskStatusInfo
from inductiva.client.models.task_step import TaskStep
from typing import Optional, Set
from typing_extensions import Self


class TaskWithStatusHistory(BaseModel):
    """
    TaskWithStatusHistory
    """

  # noqa: E501
    task_id: StrictStr
    status: TaskStatusCode
    status_alias: StrictStr
    simulator: StrictStr
    storage_path: Optional[StrictStr] = None
    storage_input_path: Optional[StrictStr] = None
    storage_output_path: Optional[StrictStr] = None
    container_image: Optional[StrictStr] = None
    project: StrictStr
    is_terminated: StrictBool
    create_time: Optional[datetime] = None
    input_submit_time: Optional[datetime] = None
    start_time: Optional[datetime] = None
    computation_start_time: Optional[datetime] = None
    computation_end_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    estimated_computation_cost: Optional[Union[StrictFloat, StrictInt]] = None
    metrics: Optional[TaskMetrics] = None
    executer: Optional[Executer] = None
    machine_group_name: Optional[StrictStr] = None
    machine_group_id: Optional[StrictStr] = None
    error_detail: Optional[StrictStr] = None
    input_resources: Optional[List[StrictStr]] = None
    stream_zip: Optional[StrictBool] = True
    num_retries: StrictInt
    compress_with: Optional[CompressionMethod] = None
    task_metadata: Optional[Dict[str, StrictStr]] = None
    extra_params: Optional[Dict[str, Any]] = None
    resubmit_on_preemption: Optional[StrictBool] = False
    duration_seconds: Optional[Union[StrictFloat, StrictInt]]
    status_history: List[TaskStatusInfo]
    machine_operations: List[TaskMachineOperation]
    steps: Optional[List[TaskStep]] = None
    __properties: ClassVar[List[str]] = [
        "task_id", "status", "status_alias", "simulator", "storage_path",
        "storage_input_path", "storage_output_path", "container_image",
        "project", "is_terminated", "create_time", "input_submit_time",
        "start_time", "computation_start_time", "computation_end_time",
        "end_time", "estimated_computation_cost", "metrics", "executer",
        "machine_group_name", "machine_group_id", "error_detail",
        "input_resources", "stream_zip", "num_retries", "compress_with",
        "task_metadata", "extra_params", "resubmit_on_preemption",
        "duration_seconds", "status_history", "machine_operations", "steps"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of TaskWithStatusHistory from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metrics
        if self.metrics:
            _dict['metrics'] = self.metrics.to_dict()
        # override the default output from pydantic by calling `to_dict()` of executer
        if self.executer:
            _dict['executer'] = self.executer.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in status_history (list)
        _items = []
        if self.status_history:
            for _item_status_history in self.status_history:
                if _item_status_history:
                    _items.append(_item_status_history.to_dict())
            _dict['status_history'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in machine_operations (list)
        _items = []
        if self.machine_operations:
            for _item_machine_operations in self.machine_operations:
                if _item_machine_operations:
                    _items.append(_item_machine_operations.to_dict())
            _dict['machine_operations'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in steps (list)
        _items = []
        if self.steps:
            for _item_steps in self.steps:
                if _item_steps:
                    _items.append(_item_steps.to_dict())
            _dict['steps'] = _items
        # set to None if storage_path (nullable) is None
        # and model_fields_set contains the field
        if self.storage_path is None and "storage_path" in self.model_fields_set:
            _dict['storage_path'] = None

        # set to None if storage_input_path (nullable) is None
        # and model_fields_set contains the field
        if self.storage_input_path is None and "storage_input_path" in self.model_fields_set:
            _dict['storage_input_path'] = None

        # set to None if storage_output_path (nullable) is None
        # and model_fields_set contains the field
        if self.storage_output_path is None and "storage_output_path" in self.model_fields_set:
            _dict['storage_output_path'] = None

        # set to None if container_image (nullable) is None
        # and model_fields_set contains the field
        if self.container_image is None and "container_image" in self.model_fields_set:
            _dict['container_image'] = None

        # set to None if create_time (nullable) is None
        # and model_fields_set contains the field
        if self.create_time is None and "create_time" in self.model_fields_set:
            _dict['create_time'] = None

        # set to None if input_submit_time (nullable) is None
        # and model_fields_set contains the field
        if self.input_submit_time is None and "input_submit_time" in self.model_fields_set:
            _dict['input_submit_time'] = None

        # set to None if start_time (nullable) is None
        # and model_fields_set contains the field
        if self.start_time is None and "start_time" in self.model_fields_set:
            _dict['start_time'] = None

        # set to None if computation_start_time (nullable) is None
        # and model_fields_set contains the field
        if self.computation_start_time is None and "computation_start_time" in self.model_fields_set:
            _dict['computation_start_time'] = None

        # set to None if computation_end_time (nullable) is None
        # and model_fields_set contains the field
        if self.computation_end_time is None and "computation_end_time" in self.model_fields_set:
            _dict['computation_end_time'] = None

        # set to None if end_time (nullable) is None
        # and model_fields_set contains the field
        if self.end_time is None and "end_time" in self.model_fields_set:
            _dict['end_time'] = None

        # set to None if estimated_computation_cost (nullable) is None
        # and model_fields_set contains the field
        if self.estimated_computation_cost is None and "estimated_computation_cost" in self.model_fields_set:
            _dict['estimated_computation_cost'] = None

        # set to None if metrics (nullable) is None
        # and model_fields_set contains the field
        if self.metrics is None and "metrics" in self.model_fields_set:
            _dict['metrics'] = None

        # set to None if executer (nullable) is None
        # and model_fields_set contains the field
        if self.executer is None and "executer" in self.model_fields_set:
            _dict['executer'] = None

        # set to None if machine_group_name (nullable) is None
        # and model_fields_set contains the field
        if self.machine_group_name is None and "machine_group_name" in self.model_fields_set:
            _dict['machine_group_name'] = None

        # set to None if machine_group_id (nullable) is None
        # and model_fields_set contains the field
        if self.machine_group_id is None and "machine_group_id" in self.model_fields_set:
            _dict['machine_group_id'] = None

        # set to None if error_detail (nullable) is None
        # and model_fields_set contains the field
        if self.error_detail is None and "error_detail" in self.model_fields_set:
            _dict['error_detail'] = None

        # set to None if task_metadata (nullable) is None
        # and model_fields_set contains the field
        if self.task_metadata is None and "task_metadata" in self.model_fields_set:
            _dict['task_metadata'] = None

        # set to None if extra_params (nullable) is None
        # and model_fields_set contains the field
        if self.extra_params is None and "extra_params" in self.model_fields_set:
            _dict['extra_params'] = None

        # set to None if duration_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.duration_seconds is None and "duration_seconds" in self.model_fields_set:
            _dict['duration_seconds'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of TaskWithStatusHistory from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "task_id":
                obj.get("task_id"),
            "status":
                obj.get("status"),
            "status_alias":
                obj.get("status_alias"),
            "simulator":
                obj.get("simulator"),
            "storage_path":
                obj.get("storage_path"),
            "storage_input_path":
                obj.get("storage_input_path"),
            "storage_output_path":
                obj.get("storage_output_path"),
            "container_image":
                obj.get("container_image"),
            "project":
                obj.get("project"),
            "is_terminated":
                obj.get("is_terminated"),
            "create_time":
                obj.get("create_time"),
            "input_submit_time":
                obj.get("input_submit_time"),
            "start_time":
                obj.get("start_time"),
            "computation_start_time":
                obj.get("computation_start_time"),
            "computation_end_time":
                obj.get("computation_end_time"),
            "end_time":
                obj.get("end_time"),
            "estimated_computation_cost":
                obj.get("estimated_computation_cost"),
            "metrics":
                TaskMetrics.from_dict(obj["metrics"])
                if obj.get("metrics") is not None else None,
            "executer":
                Executer.from_dict(obj["executer"])
                if obj.get("executer") is not None else None,
            "machine_group_name":
                obj.get("machine_group_name"),
            "machine_group_id":
                obj.get("machine_group_id"),
            "error_detail":
                obj.get("error_detail"),
            "input_resources":
                obj.get("input_resources"),
            "stream_zip":
                obj.get("stream_zip")
                if obj.get("stream_zip") is not None else True,
            "num_retries":
                obj.get("num_retries"),
            "compress_with":
                obj.get("compress_with"),
            "task_metadata":
                obj.get("task_metadata"),
            "extra_params":
                obj.get("extra_params"),
            "resubmit_on_preemption":
                obj.get("resubmit_on_preemption")
                if obj.get("resubmit_on_preemption") is not None else False,
            "duration_seconds":
                obj.get("duration_seconds"),
            "status_history": [
                TaskStatusInfo.from_dict(_item)
                for _item in obj["status_history"]
            ] if obj.get("status_history") is not None else None,
            "machine_operations": [
                TaskMachineOperation.from_dict(_item)
                for _item in obj["machine_operations"]
            ] if obj.get("machine_operations") is not None else None,
            "steps": [TaskStep.from_dict(_item) for _item in obj["steps"]]
                     if obj.get("steps") is not None else None
        })
        return _obj



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from typing import Optional, Set
from typing_extensions import Self


class UsageStatistics(BaseModel):
    """
    UsageStatistics
    """

  # noqa: E501
    period: Optional[StrictStr] = None
    total_core_hours_failed: Optional[Union[StrictFloat, StrictInt]] = None
    total_core_hours: Optional[Union[StrictFloat, StrictInt]] = None
    total_tasks: Optional[StrictInt] = None
    total_failed_tasks: Optional[StrictInt] = None
    avg_total_tasks_duration: Optional[Union[StrictFloat, StrictInt]] = None
    avg_computation_seconds_task_duration: Optional[Union[StrictFloat,
                                                          StrictInt]] = None
    success_rate: Optional[Union[StrictFloat, StrictInt]] = None
    avg_estimated_cost: Optional[Union[StrictFloat, StrictInt]] = None
    total_estimated_cost: Optional[Union[StrictFloat, StrictInt]] = None
    total_failed_cost: Optional[Union[StrictFloat, StrictInt]] = None
    __properties: ClassVar[List[str]] = [
        "period", "total_core_hours_failed", "total_core_hours", "total_tasks",
        "total_failed_tasks", "avg_total_tasks_duration",
        "avg_computation_seconds_task_duration", "success_rate",
        "avg_estimated_cost", "total_estimated_cost", "total_failed_cost"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of UsageStatistics from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if period (nullable) is None
        # and model_fields_set contains the field
        if self.period is None and "period" in self.model_fields_set:
            _dict['period'] = None

        # set to None if total_core_hours_failed (nullable) is None
        # and model_fields_set contains the field
        if self.total_core_hours_failed is None and "total_core_hours_failed" in self.model_fields_set:
            _dict['total_core_hours_failed'] = None

        # set to None if total_core_hours (nullable) is None
        # and model_fields_set contains the field
        if self.total_core_hours is None and "total_core_hours" in self.model_fields_set:
            _dict['total_core_hours'] = None

        # set to None if total_tasks (nullable) is None
        # and model_fields_set contains the field
        if self.total_tasks is None and "total_tasks" in self.model_fields_set:
            _dict['total_tasks'] = None

        # set to None if total_failed_tasks (nullable) is None
        # and model_fields_set contains the field
        if self.total_failed_tasks is None and "total_failed_tasks" in self.model_fields_set:
            _dict['total_failed_tasks'] = None

        # set to None if avg_total_tasks_duration (nullable) is None
        # and model_fields_set contains the field
        if self.avg_total_tasks_duration is None and "avg_total_tasks_duration" in self.model_fields_set:
            _dict['avg_total_tasks_duration'] = None

        # set to None if avg_computation_seconds_task_duration (nullable) is None
        # and model_fields_set contains the field
        if self.avg_computation_seconds_task_duration is None and "avg_computation_seconds_task_duration" in self.model_fields_set:
            _dict['avg_computation_seconds_task_duration'] = None

        # set to None if success_rate (nullable) is None
        # and model_fields_set contains the field
        if self.success_rate is None and "success_rate" in self.model_fields_set:
            _dict['success_rate'] = None

        # set to None if avg_estimated_cost (nullable) is None
        # and model_fields_set contains the field
        if self.avg_estimated_cost is None and "avg_estimated_cost" in self.model_fields_set:
            _dict['avg_estimated_cost'] = None

        # set to None if total_estimated_cost (nullable) is None
        # and model_fields_set contains the field
        if self.total_estimated_cost is None and "total_estimated_cost" in self.model_fields_set:
            _dict['total_estimated_cost'] = None

        # set to None if total_failed_cost (nullable) is None
        # and model_fields_set contains the field
        if self.total_failed_cost is None and "total_failed_cost" in self.model_fields_set:
            _dict['total_failed_cost'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of UsageStatistics from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "period":
                obj.get("period"),
            "total_core_hours_failed":
                obj.get("total_core_hours_failed"),
            "total_core_hours":
                obj.get("total_core_hours"),
            "total_tasks":
                obj.get("total_tasks"),
            "total_failed_tasks":
                obj.get("total_failed_tasks"),
            "avg_total_tasks_duration":
                obj.get("avg_total_tasks_duration"),
            "avg_computation_seconds_task_duration":
                obj.get("avg_computation_seconds_task_duration"),
            "success_rate":
                obj.get("success_rate"),
            "avg_estimated_cost":
                obj.get("avg_estimated_cost"),
            "total_estimated_cost":
                obj.get("total_estimated_cost"),
            "total_failed_cost":
                obj.get("total_failed_cost")
        })
        return _obj



"""API for Benchmarking"""
import enum
import json
import csv
import logging
from typing import Optional, Union
import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm
from typing_extensions import Self
from inductiva import types, resources, projects, simulators, client
from inductiva.client.models import TaskStatusCode
from inductiva.projects.project import ProjectType
from inductiva.utils.format_utils import CURRENCY_SYMBOL, TIME_UNIT

from concurrent.futures import ThreadPoolExecutor, as_completed


class ExportFormat(enum.Enum):
    """Enumeration of supported benchmark export formats."""
    JSON = "json"
    CSV = "csv"


class SelectMode(enum.Enum):
    """
    Enumeration of supported data selection modes, specifying which data
    should be included in the benchmarking results.
    """
    ALL = "all"
    DISTINCT = "distinct"


class Benchmark(projects.Project):
    """Represents the benchmark runner."""

    class InfoKey:
        """Constant keys for the info used in the benchmark runs.

        :meta private:
        """
        TASK_ID = "task_id"
        SIMULATOR = "simulator"
        MACHINE_TYPE = "machine_type"
        TIME = f"computation_time ({TIME_UNIT})"
        COST = f"estimated_computation_cost ({CURRENCY_SYMBOL})"

    def __init__(self, name: str, verbose: bool = False):
        """
        Initializes a new Benchmark instance.

        Args:
            name (str): The name of the benchmark runner.
        """
        super().__init__(name=name)
        self.runs = []
        self.simulator = None
        self.input_dir = None
        self.on = None
        self.kwargs = {}
        self.verbose = verbose

    def _get_project_type(self):
        return ProjectType.BENCHMARK

    def set_default(
        self,
        simulator: Optional[simulators.Simulator] = None,
        input_dir: Optional[str] = None,
        on: Optional[types.ComputationalResources] = None,
        **kwargs,
    ) -> Self:
        """
        Sets default parameters for the benchmark runner.

        This method allows you to configure default settings for the benchmark,
        which will be used in subsequent runs unless explicitly overridden.

        Args:
            simulator (Optional[Simulator]): The simulator instance to be used
                as the default for future runs. If not provided, the current
                simulator will remain unchanged.
            input_dir (Optional[str]): The directory path for input files. If
                not provided, the current input directory will remain unchanged.
            on (Optional[types.ComputationalResources]): The computational
                resources to use for running the simulations. If not specified,
                the current resources will remain unchanged.
            **kwargs: Additional keyword arguments to set as default parameters
                for the simulations. These will update any existing parameters
                with the same names.

        Returns:
            Self: The current instance for method chaining.
        """
        self.simulator = simulator or self.simulator
        self.input_dir = input_dir or self.input_dir
        self.on = on or self.on
        self.kwargs = {**self.kwargs, **kwargs}
        return self

    def add_run(
        self,
        simulator: Optional[simulators.Simulator] = None,
        input_dir: Optional[str] = None,
        on: Optional[types.ComputationalResources] = None,
        **kwargs,
    ) -> Self:
        """
        Adds a simulation run to the benchmark.

        Args:
            simulator (Optional[Simulator]): The simulator to be used for this
                run. If not provided, the previously set simulator will be used.
            input_dir (Optional[str]): The directory containing input files for
                the simulation. If not provided, the previously set input
                directory will be used.
            on (Optional[types.ComputationalResources]): The computational
                resources to run the simulation on. If not provided, the
                previously set resources will be used.
            **kwargs: Additional keyword arguments for the simulator run. These
                will overwrite any previously set parameters.

        Returns:
            Self: The current instance for method chaining.
        """
        self.runs.append((
            simulator or self.simulator,
            input_dir or self.input_dir,
            on or self.on,
            {
                **self.kwargs,
                **kwargs
            },
        ))
        return self

    def run(self, num_repeats: int = 2, wait_for_quotas: bool = True) -> Self:
        """
        Executes all added runs.

        Each run is executed the specified number of times, and the collection
        of runs is cleared afterwards.

        Args:
            num_repeats (int): The number of times to repeat each simulation
                run (default is 2).
            wait_for_quotas (bool): Indicates whether to wait for quotas to
                become available before starting each resource. If `True`, the
                program will actively wait in a loop, periodically sleeping and
                checking for quotas. If `False`, the program crashes if quotas
                are not available (default is `True`).

        Returns:
            Self: The current instance for method chaining.
        """
        if not self.verbose:
            logging.info(
                "Preparing tasks for Benchmark to start. "
                "This may take a few minutes.\n"
                "Note that stopping this process will \033[1minterrupt\033[0m "
                "the submission of the tasks. Please wait...\n")
        for simulator, input_dir, machine_group, kwargs in self.runs:
            if not machine_group.started:
                machine_group.start(wait_for_quotas=wait_for_quotas,
                                    verbose=self.verbose)
            for _ in range(num_repeats):
                simulator.run(input_dir=input_dir,
                              on=machine_group,
                              project=self.name,
                              resubmit_on_preemption=True,
                              verbose=self.verbose,
                              **kwargs)
        self.runs.clear()
        logging.info(
            "■ Benchmark \033[1m%s\033[0m has started.\n"
            "  Go to https://console.inductiva.ai/benchmarks/%s "
            "for more details.\n", self.name, self.name)
        return self

    def wait(self) -> Self:
        """
        Waits for all running tasks to complete.

        Returns:
            Self: The current instance for method chaining.
        """
        tasks = self.get_tasks()

        completed_tasks = [task for task in tasks if task.info.is_terminal]
        running_tasks = [task for task in tasks if not task.info.is_terminal]

        logging.info("Waiting for Benchmark \033[1m%s\033[0m to complete...\n",
                     self.name)

        with tqdm.tqdm(total=len(tasks),
                       desc="Running Benchmark",
                       bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} {unit}",
                       initial=len(completed_tasks),
                       unit="tasks") as pbar:
            with ThreadPoolExecutor() as executor:
                future_to_task = {
                    executor.submit(
                        lambda t: t.wait(download_std_on_completion=False,
                                         silent_mode=True), task):
                        task for task in running_tasks
                }

                for future in as_completed(future_to_task):
                    with logging_redirect_tqdm():
                        task = future_to_task[future]
                        status = future.result()

                        if status != TaskStatusCode.SUCCESS:
                            logging.info(
                                "Task %s completed with status: %s\n"
                                "   · To understand why the task did not "
                                "complete successfully go to "
                                "https://console.inductiva.ai/tasks/%s",
                                task.id, status, task.id)

                        # Update progress bar for each completed task
                        pbar.update(1)

        logging.info("\n")
        return self

    def export(
        self,
        fmt: Union[ExportFormat, str] = ExportFormat.JSON,
        filename: Optional[str] = None,
        status: Optional[Union[TaskStatusCode, str]] = None,
        select: Union[SelectMode, str] = SelectMode.DISTINCT,
    ):
        """
        Exports the benchmark performance metrics in the specified format.

        Args:
            fmt (Union[ExportFormat, str]): The format to export the results
                in. Defaults to ExportFormat.JSON.
            filename (Optional[str]): The name of the output file to save the
                exported results. Defaults to the benchmark's name if not
                provided.
            status (Optional[Union[TaskStatusCode, str]]): The status of the
                tasks to include in the benchmarking results. Defaults to None,
                which includes all tasks.
            select (Union[SelectMode, str]): The data to include in
                the benchmarking results. Defaults to SelectMode.DISTINCT that
                includes only the parameters that vary between different runs.
        """
        if isinstance(fmt, str):
            fmt = ExportFormat[fmt.upper()]

        if isinstance(select, str):
            select = SelectMode[select.upper()]

        if status is not None:
            status = TaskStatusCode(status)

        response = self._api.get_tasks_info_without_preload_content(
            name=self.name,
            select=select,
            status=status,
        )
        info = json.loads(response.data)

        if not filename:
            filename = f"{self.name}.{fmt.value}"
        elif not filename.endswith(f".{fmt.value}"):
            filename = f"{filename}.{fmt.value}"

        if fmt == ExportFormat.JSON:
            with open(filename, mode="w", encoding="utf-8") as file:
                json_content = json.dumps(obj=info, indent=4)
                file.write(json_content)
        elif fmt == ExportFormat.CSV:
            with open(filename, mode="w", encoding="utf-8") as file:
                fieldnames = sorted(info[0].keys()) if info else []
                writer = csv.DictWriter(file, fieldnames=fieldnames)
                writer.writeheader()
                writer.writerows(info)
        else:
            raise ValueError(f"Unsupported export format: {fmt}")

    def terminate(self) -> Self:
        """
        Terminates all active machine groups associated with the
        benchmark.

        Returns:
            Self: The current instance for method chaining.
        """

        def _handle_suffix(executer):
            if executer.host_type == resources.utils.ProviderType.GCP:
                return "-".join(executer.vm_name.split("-")[:-1])
            return executer.vm_name

        machine_names = {
            _handle_suffix(task.info.executer)
            for task in self.get_tasks()
            if task.info.executer
        }

        for machine in resources.get():
            if machine.name not in machine_names:
                continue
            try:
                machine.terminate(verbose=False)
            except client.ApiException as api_exception:
                logging.warning(api_exception)

        return self



"""File Tracker module for connecting to a running task via WebRTC."""
import asyncio
import json
import uuid
import enum
import logging
import warnings
import inductiva.client
from inductiva import constants
import inductiva.client.models
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import aiortc

# STUN/TURN server configuration
ICE_SERVERS = [
    aiortc.RTCIceServer("stun:" + constants.TURN_SERVER_URL),
    aiortc.RTCIceServer("turn:" + constants.TURN_SERVER_URL)
]

aiortc_logger = logging.getLogger("aioice")
aiortc_logger.setLevel(logging.WARNING)


class Operations(enum.Enum):
    LIST = "ls"
    TAIL = "tail"
    TOP = "top"
    LAST_MODIFIED_FILE = "last_modified_file"


class FileTracker:
    """File Tracker class for connecting to a running task via WebRTC."""

    def __init__(self):
        self.peer_connections = []

    async def setup_channel(self, pc, operation, follow=False, **kwargs):
        channel = pc.createDataChannel("file_transfer")
        queue = asyncio.Queue()
        end_event = asyncio.Event()

        @channel.on("open")
        def on_open():
            request = {"type": operation.value, "follow": follow}
            request["args"] = kwargs
            channel.send(json.dumps(request))

        @channel.on("message")
        async def on_message(message):
            await queue.put(json.loads(message))
            if not follow:
                end_event.set()

        @channel.on("close")
        async def on_close():
            await queue.put(None)

        return queue, end_event

    def create_peer_connection(self):
        pc = aiortc.RTCPeerConnection(
            aiortc.RTCConfiguration(iceServers=ICE_SERVERS))
        self.peer_connections.append(pc)
        return pc

    async def connect_to_task(
        self,
        api: inductiva.client.TasksApi,
        pc,
        task_id,
    ):
        connection_id = str(uuid.uuid4())
        api.register_task(
            task_connection=inductiva.client.models.TaskConnection(
                sender_id=connection_id),
            task_id=task_id,
        )

        offer = await pc.createOffer()
        await pc.setLocalDescription(offer)

        api.offer_task(
            task_connection=inductiva.client.models.TaskConnection(
                sender_id=connection_id,
                receiver_id=task_id,
                type="offer",
                sdp=pc.localDescription.sdp,
            ),
            task_id=task_id,
        )

        resp = api.get_message_without_preload_content(client=connection_id,
                                                       task_id=task_id)

        if resp.status == 204:
            return False

        data = json.loads(resp.data)
        if data["type"] == "answer":
            await pc.setRemoteDescription(
                aiortc.RTCSessionDescription(sdp=data["sdp"],
                                             type=data["type"]))

        return True

    async def cleanup(self):
        for pc in self.peer_connections:
            await pc.close()



"""OpenFAST module of the API for wind turbine simulations."""
from typing import Optional, Union

from inductiva import simulators, tasks, types


class OpenFAST(simulators.Simulator):
    """Class to invoke a generic OpenFAST simulation on the API.

    """

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the OpenFAST simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "openfast"

    def run(self,
            input_dir: Optional[str],
            commands: types.Commands,
            *,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            commands: List of commands to run using the OpenFAST simulator.
            on: The computational resource to launch the simulation on.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Delft3D module of the API."""
from typing import Optional, Union

from inductiva import simulators, types


class Delft3D(simulators.Simulator):
    """Class to invoke a generic Delft3D simulation on the API."""

    _default_delft3d_device = "cpu"

    def __init__(
        self,
        /,
        version: Optional[str] = None,
        use_dev: bool = False,
    ):
        super().__init__(version, use_dev, self._default_delft3d_device)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "delft3d"

    def run(self,
            input_dir: Optional[str],
            *,
            commands: Optional[types.Commands] = None,
            shell_script: Optional[str] = None,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs):

        if commands is None and shell_script is None:
            raise ValueError("Either commands or shell_script "
                             "must be provided.")
        if commands is not None and shell_script is not None:
            raise ValueError("Only one of commands or shell_script "
                             "can be provided.")

        if shell_script is not None:
            #Check if the shell script exists

            self._input_files_exist(input_dir=input_dir,
                                    remote_assets=remote_assets,
                                    shell_script=shell_script)
            commands = [f"bash {shell_script}"]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from typing import Optional, Set
from typing_extensions import Self


class MachineInfo(BaseModel):
    """
    MachineInfo
    """

  # noqa: E501
    uuid: StrictStr
    cpu_count_logical: StrictInt
    cpu_count_physical: StrictInt
    memory_bytes: StrictInt
    host_name: Optional[StrictStr] = None
    started_at: Optional[datetime] = None
    terminated_at: Optional[datetime] = None
    last_seen_at: Optional[datetime] = None
    current_task_id: Optional[StrictStr] = None
    disk_size_gb: Optional[Union[StrictFloat, StrictInt]] = None
    __properties: ClassVar[List[str]] = [
        "uuid", "cpu_count_logical", "cpu_count_physical", "memory_bytes",
        "host_name", "started_at", "terminated_at", "last_seen_at",
        "current_task_id", "disk_size_gb"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of MachineInfo from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if host_name (nullable) is None
        # and model_fields_set contains the field
        if self.host_name is None and "host_name" in self.model_fields_set:
            _dict['host_name'] = None

        # set to None if started_at (nullable) is None
        # and model_fields_set contains the field
        if self.started_at is None and "started_at" in self.model_fields_set:
            _dict['started_at'] = None

        # set to None if terminated_at (nullable) is None
        # and model_fields_set contains the field
        if self.terminated_at is None and "terminated_at" in self.model_fields_set:
            _dict['terminated_at'] = None

        # set to None if last_seen_at (nullable) is None
        # and model_fields_set contains the field
        if self.last_seen_at is None and "last_seen_at" in self.model_fields_set:
            _dict['last_seen_at'] = None

        # set to None if current_task_id (nullable) is None
        # and model_fields_set contains the field
        if self.current_task_id is None and "current_task_id" in self.model_fields_set:
            _dict['current_task_id'] = None

        # set to None if disk_size_gb (nullable) is None
        # and model_fields_set contains the field
        if self.disk_size_gb is None and "disk_size_gb" in self.model_fields_set:
            _dict['disk_size_gb'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of MachineInfo from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "uuid": obj.get("uuid"),
            "cpu_count_logical": obj.get("cpu_count_logical"),
            "cpu_count_physical": obj.get("cpu_count_physical"),
            "memory_bytes": obj.get("memory_bytes"),
            "host_name": obj.get("host_name"),
            "started_at": obj.get("started_at"),
            "terminated_at": obj.get("terminated_at"),
            "last_seen_at": obj.get("last_seen_at"),
            "current_task_id": obj.get("current_task_id"),
            "disk_size_gb": obj.get("disk_size_gb")
        })
        return _obj



"""Register CLI command for login."""
from typing import List
from tqdm import tqdm
import argparse
import requests
import getpass
import pkgutil
import os

import inductiva
from inductiva import constants, users, utils


def download_example_scripts() -> List[str]:
    """Downloads example scripts from the Inductiva Github repository.

    The scripts are downloaded to the `inductiva_examples` directory.

    Returns:
        list: A list of the names of the downloaded example scripts.
    """
    modules = [
        name
        for _, name, _ in pkgutil.iter_modules(inductiva.simulators.__path__)
    ]
    #openfoam as a different naming convention
    modules.append("openfoam_esi")
    modules.append("openfoam_foundation")

    os.makedirs("inductiva_examples", exist_ok=True)
    downloaded = []
    for module in tqdm(modules,
                       leave=False,
                       desc="Downloading example scripts",
                       unit="script"):
        url = constants.INDUCTIVA_GIT_EXAMPLES_URL + f"{module}/{module}.py"

        response = requests.get(url, timeout=10)
        if response.status_code == 200:
            downloaded.append(f"{module}.py")
            with open(f"inductiva_examples/{module}.py", "w",
                      encoding="utf-8") as f:
                f.write(response.text)
    return downloaded


def login(args):
    """Prompts the user to enter their API Key and stores it securely."""

    # pylint: disable=trailing-whitespace,line-too-long
    inductiva_art = r"""     ___  _   _  ____   _   _   ____  _____  ___ __     __ _
    |_ _|| \ | ||  _ \ | | | | / ___||_   _||_ _|\ \   / // \
     | | |  \| || | | || | | || |      | |   | |  \ \ / // _ \
     | | | |\  || |_| || |_| || |___   | |   | |   \ V // ___ \
    |___||_| \_||____/  \___/  \____|  |_|  |___|   \_//_/   \_\
    """
    print(inductiva_art)
    if os.path.exists(constants.API_KEY_FILE_PATH):
        print(
            "    You are already logged in. Run `inductiva auth logout` if you "
            "want to log out. \n"
            "    Setting a new API Key will erase the existing one.")

    prompt_func = getpass.getpass if args.private else input
    warning = " (input will not be visible)" if args.private else ""

    prompt = prompt_func(
        "    To log in, you need an API Key. You can obtain it "
        "from your account at https://console.inductiva.ai/account.\n"
        f"Please paste your API Key here{warning}: ")

    api_key = prompt.strip()

    # Set the API Key
    inductiva.set_api_key(api_key, login_message=False)

    # If the API Key is invalid, this will raise an exception
    user_info = users.get_info()

    first_log_in = utils.set_stored_api_key(api_key)

    user_name = user_info.name or ""

    if first_log_in:
        print(f"Welcome back {user_name}!")
    else:
        print("\n")
        print(f" ■ Welcome {user_name}!\n"
              "Since this is your first time logging in, we will download "
              "some example scripts for you to get started.\n")

        downloaded = download_example_scripts()

        print("The examples are located in the `inductiva_examples` folder.\n"
              "You can run them using the command "
              "`python inductiva_examples/<example.py>`.\n")
        print("Available examples:")
        for i in range(0, len(downloaded), 3):
            line = [word.ljust(25) for word in downloaded[i:i + 3]]
            print("".join(line))
        print("\nRun your first simulation with "
              "`python inductiva_examples/openfoam_esi.py`\n")


def register(parser):
    """Register the login command."""
    subparser = parser.add_parser("login",
                                  help="Login using Inductiva API Key.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva login` command allows you to log in using your key.\n"
        "You can obtain your API Key from your account at "
        "https://console.inductiva.ai/account.\n")

    subparser.add_argument("--private",
                           action="store_true",
                           help="Hide API Key.")

    subparser.set_defaults(func=login)
